/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.server.metrics.prometheus;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.errorprone.annotations.CanIgnoreReturnValue;
import com.google.errorprone.annotations.Var;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import javax.annotation.CheckReturnValue;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import javax.annotation.concurrent.Immutable;
import javax.annotation.concurrent.NotThreadSafe;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;
import org.immutables.value.Generated;

/**
 * Immutable implementation of {@link QueryResult.Data.Result}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new QueryResult.Data.Result.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableResult.of()}.
 */
@Generated(from = "QueryResult.Data.Result", generator = "Immutables")
@SuppressWarnings({"all"})
@SuppressFBWarnings
@ParametersAreNonnullByDefault
@javax.annotation.Generated("org.immutables.processor.ProxyProcessor")
@Immutable
@CheckReturnValue
final class ImmutableResult implements QueryResult.Data.Result {
  private final Map<String, String> metric;
  private final List<Object> value;

  private ImmutableResult(
      Map<String, ? extends String> metric,
      Iterable<? extends Object> value) {
    this.metric = createUnmodifiableMap(true, false, metric);
    this.value = createUnmodifiableList(false, createSafeList(value, true, false));
  }

  private ImmutableResult(
      ImmutableResult original,
      Map<String, String> metric,
      List<Object> value) {
    this.metric = metric;
    this.value = value;
  }

  /**
   * @return The value of the {@code metric} attribute
   */
  @JsonProperty("metric")
  @Override
  public Map<String, String> getMetric() {
    return metric;
  }

  /**
   * @return The value of the {@code value} attribute
   */
  @JsonProperty("value")
  @Override
  public List<Object> getValue() {
    return value;
  }

  /**
   * Copy the current immutable object by replacing the {@link QueryResult.Data.Result#getMetric() metric} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the metric map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableResult withMetric(Map<String, ? extends String> entries) {
    if (this.metric == entries) return this;
    Map<String, String> newValue = createUnmodifiableMap(true, false, entries);
    return validate(new ImmutableResult(this, newValue, this.value));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link QueryResult.Data.Result#getValue() value}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableResult withValue(Object... elements) {
    List<Object> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableResult(this, this.metric, newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link QueryResult.Data.Result#getValue() value}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of value elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableResult withValue(Iterable<? extends Object> elements) {
    if (this.value == elements) return this;
    List<Object> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableResult(this, this.metric, newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableResult} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(@Nullable Object another) {
    if (this == another) return true;
    return another instanceof ImmutableResult
        && equalTo((ImmutableResult) another);
  }

  private boolean equalTo(ImmutableResult another) {
    return metric.equals(another.metric)
        && value.equals(another.value);
  }

  /**
   * Computes a hash code from attributes: {@code metric}, {@code value}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    @Var int h = 5381;
    h += (h << 5) + metric.hashCode();
    h += (h << 5) + value.hashCode();
    return h;
  }


  /**
   * Prints the immutable value {@code Result} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Result{"
        + "metric=" + metric
        + ", value=" + value
        + "}";
  }

  /**
   * Construct a new immutable {@code Result} instance.
   * @param metric The value for the {@code metric} attribute
   * @param value The value for the {@code value} attribute
   * @return An immutable Result instance
   */
  public static QueryResult.Data.Result of(Map<String, String> metric, List<Object> value) {
    return of(metric, (Iterable<? extends Object>) value);
  }

  /**
   * Construct a new immutable {@code Result} instance.
   * @param metric The value for the {@code metric} attribute
   * @param value The value for the {@code value} attribute
   * @return An immutable Result instance
   */
  public static QueryResult.Data.Result of(Map<String, ? extends String> metric, Iterable<? extends Object> value) {
    return validate(new ImmutableResult(metric, value));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableResult validate(ImmutableResult instance) {
    Set<ConstraintViolation<ImmutableResult>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link QueryResult.Data.Result} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Result instance
   */
  public static QueryResult.Data.Result copyOf(QueryResult.Data.Result instance) {
    if (instance instanceof ImmutableResult) {
      return (ImmutableResult) instance;
    }
    return new QueryResult.Data.Result.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link QueryResult.Data.Result Result}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @Generated(from = "QueryResult.Data.Result", generator = "Immutables")
  @NotThreadSafe
  public static class Builder {
    private Map<String, String> metric = new LinkedHashMap<String, String>();
    private List<Object> value = new ArrayList<Object>();

    /**
     * Creates a builder for {@link QueryResult.Data.Result Result} instances.
     * <pre>
     * new QueryResult.Data.Result.Builder()
     *    .putMetric|putAllMetric(String =&gt; String) // {@link QueryResult.Data.Result#getMetric() metric} mappings
     *    .addValue|addAllValue(Object) // {@link QueryResult.Data.Result#getValue() value} elements
     *    .build();
     * </pre>
     */
    public Builder() {
      if (!(this instanceof QueryResult.Data.Result.Builder)) {
        throw new UnsupportedOperationException("Use: new QueryResult.Data.Result.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code Result} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Result.Builder createFrom(QueryResult.Data.Result instance) {
      Objects.requireNonNull(instance, "instance");
      putAllMetric(instance.getMetric());
      addAllValue(instance.getValue());
      return (QueryResult.Data.Result.Builder) this;
    }

    /**
     * Put one entry to the {@link QueryResult.Data.Result#getMetric() metric} map.
     * @param key The key in the metric map
     * @param value The associated value in the metric map
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Result.Builder putMetric(String key, String value) {
      this.metric.put(
          Objects.requireNonNull(key, "metric key"),
          Objects.requireNonNull(value, "metric value"));
      return (QueryResult.Data.Result.Builder) this;
    }

    /**
     * Put one entry to the {@link QueryResult.Data.Result#getMetric() metric} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Result.Builder putMetric(Map.Entry<String, ? extends String> entry) {
      String k = entry.getKey();
      String v = entry.getValue();
      this.metric.put(
          Objects.requireNonNull(k, "metric key"),
          Objects.requireNonNull(v, "metric value"));
      return (QueryResult.Data.Result.Builder) this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link QueryResult.Data.Result#getMetric() metric} map. Nulls are not permitted
     * @param entries The entries that will be added to the metric map
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("metric")
    public final QueryResult.Data.Result.Builder metric(Map<String, ? extends String> entries) {
      this.metric.clear();
      return putAllMetric(entries);
    }

    /**
     * Put all mappings from the specified map as entries to {@link QueryResult.Data.Result#getMetric() metric} map. Nulls are not permitted
     * @param entries The entries that will be added to the metric map
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Result.Builder putAllMetric(Map<String, ? extends String> entries) {
      for (Map.Entry<String, ? extends String> e : entries.entrySet()) {
        String k = e.getKey();
        String v = e.getValue();
        this.metric.put(
            Objects.requireNonNull(k, "metric key"),
            Objects.requireNonNull(v, "metric value"));
      }
      return (QueryResult.Data.Result.Builder) this;
    }

    /**
     * Adds one element to {@link QueryResult.Data.Result#getValue() value} list.
     * @param element A value element
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Result.Builder addValue(Object element) {
      this.value.add(Objects.requireNonNull(element, "value element"));
      return (QueryResult.Data.Result.Builder) this;
    }

    /**
     * Adds elements to {@link QueryResult.Data.Result#getValue() value} list.
     * @param elements An array of value elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Result.Builder addValue(Object... elements) {
      for (Object element : elements) {
        this.value.add(Objects.requireNonNull(element, "value element"));
      }
      return (QueryResult.Data.Result.Builder) this;
    }


    /**
     * Sets or replaces all elements for {@link QueryResult.Data.Result#getValue() value} list.
     * @param elements An iterable of value elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("value")
    public final QueryResult.Data.Result.Builder value(Iterable<? extends Object> elements) {
      this.value.clear();
      return addAllValue(elements);
    }

    /**
     * Adds elements to {@link QueryResult.Data.Result#getValue() value} list.
     * @param elements An iterable of value elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Result.Builder addAllValue(Iterable<? extends Object> elements) {
      for (Object element : elements) {
        this.value.add(Objects.requireNonNull(element, "value element"));
      }
      return (QueryResult.Data.Result.Builder) this;
    }

    /**
     * Builds a new {@link QueryResult.Data.Result Result}.
     * @return An immutable instance of Result
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public QueryResult.Data.Result build() {
      return ImmutableResult.validate(new ImmutableResult(null, createUnmodifiableMap(false, false, metric), createUnmodifiableList(true, value)));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }

  private static <K, V> Map<K, V> createUnmodifiableMap(boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    switch (map.size()) {
    case 0: return Collections.emptyMap();
    case 1: {
      Map.Entry<? extends K, ? extends V> e = map.entrySet().iterator().next();
      K k = e.getKey();
      V v = e.getValue();
      if (checkNulls) {
        Objects.requireNonNull(k, "key");
        Objects.requireNonNull(v, "value");
      }
      if (skipNulls && (k == null || v == null)) {
        return Collections.emptyMap();
      }
      return Collections.singletonMap(k, v);
    }
    default: {
      Map<K, V> linkedMap = new LinkedHashMap<>(map.size());
      if (skipNulls || checkNulls) {
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
          K k = e.getKey();
          V v = e.getValue();
          if (skipNulls) {
            if (k == null || v == null) continue;
          } else if (checkNulls) {
            Objects.requireNonNull(k, "key");
            Objects.requireNonNull(v, "value");
          }
          linkedMap.put(k, v);
        }
      } else {
        linkedMap.putAll(map);
      }
      return Collections.unmodifiableMap(linkedMap);
    }
    }
  }
}
